/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.http.dfault.util;

import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.ext.connector.db.DataSourceConnector;

public class HTTPBindingProviderConfiguration {

    // initialize all instance vars with reasonable defaults

    private boolean embedStackTraceInFaults = true;
    private Map<QName, Integer> requestorTimeout;
    private int maxPayloadProcessors = 5;
    private String dataSourceConnectorName = DataSourceConnector.SYSTEM_DATA_SOURCE;
   
    /**
     * Gets amount of time before a blocking request times out, i.e the amount of time provider
     * will spend waiting on an anonymous response before throwing a fault.
     * 
     * @return value in milliseconds or -1 to indicate no timeout.
     */
    public int getRequestorTimeout(QName operationName) {
        int answer = 60000; // default to 1 minute
        if (requestorTimeout != null) {
            Integer integer = requestorTimeout.get(operationName);
            if (integer != null) {
                answer = integer.intValue();
            }
        }
        return answer;
    }

    public void setRequestorTimeout(Map<QName, Integer> requestorTimeout) {
        this.requestorTimeout = requestorTimeout;
    }
    
    /**
     * Defines DataSourceConnector name which MAY be required by binding provider to implement
     * message and/or header persistence. If element is undefined, the connector
     * 'SystemDataSource' is implied.
     * 
     * @return
     */
    public String getDataSourceConnectorName() {
        return dataSourceConnectorName;
    }

    public void setDataSourceConnectorName(String dataSourceConnectorName) {
        this.dataSourceConnectorName = dataSourceConnectorName;
    }

    /**
     * Defines maximum number of concurrent processors which process message payload, i.e.
     * which process action by binding it according to wsdl description and then passing to
     * engine.
     * 
     * @return
     */
    public int getMaxPayloadProcessors() {
        return maxPayloadProcessors;
    }

    public void setMaxPayloadProcessors(int maxPayloadProcessors) {
        this.maxPayloadProcessors = maxPayloadProcessors;
    }

    /**
     * Constrains fault handling.
     * 
     * @return true if stack trace embedded in detail. false if not.
     */
    public boolean isEmbeddStackTraceInFaults() {
        return embedStackTraceInFaults;
    }

    public void setEmbeddStackTraceInFaults(boolean embedStackTraceInFaults) {
        this.embedStackTraceInFaults = embedStackTraceInFaults;
    }

}
